from .func_gmic import *
from ..utils import get_addon_prefs, open_folder
import bpy
from os.path import join, basename, exists, dirname, abspath, splitext

'''#decorator mod
def with_renderfile(filepath):
    def with_renderfile_decorator(func):
        def decorator(*args, **kwargs):
            r = bpy.context.scene.render
            old_filepath, r.filepath = r.filepath, filepath
            try:
                func(*args, **kwargs)
            finally:
                r.filepath = old_filepath

        return decorator

    return with_renderfile_decorator

@with_renderfile("//myfile")
'''

# self with implementation
def render_filepath(filepath):
    class RenderFileRestorer:

        def __enter__(self):
            bpy.context.scene.render.film_transparent = True
            bpy.context.scene.render.filepath = filepath
            bpy.context.scene.render.resolution_percentage = bpy.context.scene.gpcolor_props.res_percentage

        def __exit__(self, type, value, traceback):
            bpy.context.scene.render.filepath = old_filepath
            bpy.context.scene.render.film_transparent = transparent
            bpy.context.scene.render.resolution_percentage = old_res

    transparent = bpy.context.scene.render.film_transparent
    old_filepath = bpy.context.scene.render.filepath
    old_res = bpy.context.scene.render.resolution_percentage
    return RenderFileRestorer()


def layer_state(gp_data):
    class LayerStateRestorer:

        def __enter__(self):
            # mask/restore other GP object ?
            self.layers_state = {l:l.hide for l in gp_data.layers}

        def __exit__(self, type, value, traceback):
            for k, v in self.layers_state.items():
                k.hide = v

    return LayerStateRestorer()

def cursor_state():
    class CursorStateRestorer:

        def __enter__(self):
            ... 
        def __exit__(self, type, value, traceback):
            bpy.context.window_manager.progress_end()

    return CursorStateRestorer()

def imtool_fp(dest='tool',name=None):
    if name:
        return join(bpy.path.abspath('//'), dest, name)
    return join(bpy.path.abspath('//'), dest)

def generate_seq_placeholder():
    with cursor_state():
        wm = bpy.context.window_manager
        wm.progress_begin(bpy.context.scene.frame_start, bpy.context.scene.frame_end)
        for i in range(bpy.context.scene.frame_start,bpy.context.scene.frame_end+1):
            fp = imtool_fp(name = f'colotmp_{str(i).zfill(4)}.png')
            wm.progress_update(i)#remap from frame range to 0-1000 with transfer_value(Value, OldMin, OldMax, NewMin, NewMax)
            if not exists(fp):
                generate_empty_image(fp)

def set_bg_img_settings(bgimg):
    bgimg.display_depth = 'FRONT' if bpy.data.version[1] < 83 else 'BACK'
    bgimg.alpha = 0.6
    bgimg.frame_method = 'FIT'#'STRETCH'


def load_bg_image(colo_fp):
    # load new image in camera background
    cam = bpy.context.scene.camera.data
    cam.show_background_images = True
    
    colo = basename(colo_fp)
    colo_img = bpy.data.images.get(colo)
    
    # load as images
    if colo_img:
        colo_img.reload()
    else:
        if not exists(colo_fp):return
        colo_img = bpy.data.images.load(colo_fp)
    
    bgimg = None
    for bg in cam.background_images:
        if bg.image == colo_img:
            bgimg = bg
            break
    
    if not bgimg:
        bgimg = cam.background_images.new()
        bgimg.image = colo_img

    set_bg_img_settings(bgimg)
    return bgimg

def load_bg_movieclip():
    # load new image in camera background
    cam = bpy.context.scene.camera.data
    cam.show_background_images = True
    
    ## load first image
    colo = f'colotmp_{str(bpy.context.scene.frame_start).zfill(4)}.png'
    colo_fp = join(imtool_fp(), colo)
    if not exists(colo_fp):
        try:
            generate_empty_image(colo_fp)
            pass
            # TODO generate empty alpha image with GMIC or fast lib (numpy ?)   
        except Exception as identifier:
            print('In load_bg_movieclip')
            print(f'NOT FOUND : {colo_fp}')
            return 
        return

    # load as movie clip
    colo_img = bpy.data.movieclips.get(colo)
    if colo_img:
        pass# colo_img.reload()# video has no reload prop
        #TODO find a way to trigger refresh automagically !!
    else:
        if not exists(colo_fp):return
        colo_img = bpy.data.movieclips.load(colo_fp)

    bgimg = None
    for bg in cam.background_images:
        if bg.clip == colo_img:
            bgimg = bg
            break

    if not bgimg:
        bgimg = cam.background_images.new()
        bgimg.source = 'MOVIE_CLIP'
        bgimg.clip = colo_img

    set_bg_img_settings(bgimg)
    return bgimg


def guide_color(anim=False):
    '''render lines and spots separately > gmic > feeback into cam background'''
    scene = bpy.context.scene
    #### other solution
    ### how about conerting polylines to stroke under the hood with opencv in camera space according to layer
    ### then feed generated array to gmic, might be faster (but dont take stroke thickness into account... problem or feature ?)
    
    # Generate temporary (local for speed) folder or render in /tmp if not specified ?
    if not bpy.context.object or bpy.context.object.type != 'GPENCIL': return 1
    
    gp = bpy.context.object.data
    frame = str(scene.frame_current).zfill(4)
    line = f"line_{frame}.png"
    spot = f"spot_{frame}.png"
    colo = f"colotmp_{frame}.png"
    with layer_state(gp):
        # show/hide layers to render lines/spots only
        for l in gp.layers:
            l.hide = any(x in l.info for x in ('spot', 'colo'))#keep only lines
        # better hide by material namespace ?
            
        with render_filepath(f"//tool/{line}"):
            bpy.ops.render.render(animation = anim, write_still=True)

        # show/hide layers to render spots only
        for l in gp.layers:
            l.hide = 'spot' not in l.info#keep only spots
        
        with render_filepath(f"//tool/{spot}"):
            bpy.ops.render.render(animation = anim, write_still=True)
    line_fp = join(imtool_fp(), line)
    spot_fp = join(imtool_fp(), spot)
    colo_fp = join(imtool_fp(), colo)
    
    propagate_color(line_fp, spot_fp, colo_fp)
    ## ~4.6sec


    # now try and check if openCV or gmic can smooth vectorize the stuff

    ## ! surely possible to avoid writting the file like giving it back as a numpy array and keep it within blender !
    ## >> just feed numpy array to gmic and get output ?
    ## Again maybe possible to avoid writing to disk but more complicated to send to gmic (use slot 8 and 9 ?)

    # clear line/spot render if necessary...


class GMICOLOR_OT_propagate_spots(bpy.types.Operator):
    """
    Propagate the spots with a gmic call
    use shift+clic to force reload after operation.
    """
    bl_idname = "bgmic.propagate_color"
    bl_label = "Gmic propagate color"

    @classmethod
    def poll(cls, context):
        return context.active_object is not None and context.active_object.type == 'GPENCIL'

    anim : bpy.props.BoolProperty(
        name="animation", description="render and propagate color for whole animation", default=False, subtype='NONE', options={'ANIMATABLE'})#HIDDEN

    ## TODO set preview mode (with low percentage)

    # subtype (string) – Enumerator in ['FILE_PATH', 'DIR_PATH', 'FILE_NAME', 'BYTE_STRING', 'PASSWORD', 'NONE'].
    # options (set) – Enumerator in ['HIDDEN', 'SKIP_SAVE', 'ANIMATABLE', 'LIBRARY_EDITABLE', 'PROPORTIONAL','TEXTEDIT_UPDATE'].
    mode : bpy.props.StringProperty(
        name="mode", description="Set mode for operator", default="render", maxlen=0, subtype='NONE', options={'ANIMATABLE'})

    load : bpy.props.BoolProperty(default=False)

    def execute(self, context):
        ## TODO make a warning if animation mode is on and there is a consequent number of frame with big resolutions...
        frame = str(context.scene.frame_current).zfill(4)
        colo = f'colotmp_{frame}.png'
        colo_fp = join(imtool_fp(), colo)
        bgimg = None
        
        ## render one image or full-anim
        if self.mode == 'render':
            ## Or maybe use the technique of BG rendering... with subprocess call
            bgimg = guide_color(self.anim)


        ## re-load 
        if self.mode == 'load' or self.load:
            """ if not exists(colo_fp):
                print(f'/!\\ {frame}: color was not generated')
                return {'CANCELLED'}  """
            bpy.ops.bgmic.clear_cam_bg_images(real_clear=False)#clear before load
            if self.anim:
                bgimg = load_bg_movieclip()
                generate_seq_placeholder()#generate all placeholders
            else:
                bgimg = load_bg_image(colo_fp)

            if not bgimg:
                mess = f"Not Found, Loading {'anim' if self.anim else 'image'} has failed at {dirname(colo_fp)}"
                self.report({'ERROR'}, mess)#WARNING, ERROR
                return {'CANCELLED'} 

        return {'FINISHED'}

    def invoke(self, context, event):
        self.load = event.shift#if shift is pressed, force inclusion of load
        return self.execute(context)

class GMICOLOR_OT_clear_cam_bg_images(bpy.types.Operator):
    """
    Disable and clear background images from scene camera (mask non spot imgs)
    Shift+clic for a real clear (delete all refs images from camera bg images)
    """
    bl_idname = "bgmic.clear_cam_bg_images"
    bl_label = "Clear camera background color images"

    @classmethod
    def poll(cls, context):
        return context.scene.camera

    real_clear : bpy.props.BoolProperty(default=False)

    def execute(self, context):
        # TODO be selective to user cam (filter on name)
        # TODO need to delete only "spot" name instead of full clear
        if self.real_clear:
            context.scene.camera.data.background_images.clear()
        else:
            to_del = []
            for bgimg in context.scene.camera.data.background_images:
                if bgimg.source == 'MOVIE_CLIP':
                    if not bgimg.clip or 'spot' in bgimg.clip.name or 'colotmp' in bgimg.clip.name:
                        to_del.append(bgimg)
                        continue
                    else:#just hide
                        bgimg.show_background_image = False
                else:
                    if not bgimg.image or 'spot' in bgimg.image.name or 'colotmp' in bgimg.image.name:
                        to_del.append(bgimg)
                        continue
                    else:#just hide
                        bgimg.show_background_image = False

            for bimg in reversed(to_del):
                context.scene.camera.data.background_images.remove(bimg)


        context.scene.camera.data.show_background_images = False#toggle off then on to force refresh
        # context.scene.camera.data.show_background_images = True
        return {'FINISHED'}
    
    def invoke(self, context, event):
        self.real_clear = event.shift#if shift is pressed, force inclusion of load
        return self.execute(context)

class GMICOLOR_OT_open_gmic_tool_folder(bpy.types.Operator):
    """
    Disable and clear background images from scene camera (mask non spot imgs)
    Shift+clic for a real clear (delete all refs images from camera bg images)
    """
    bl_idname = "bgmic.open_gmic_tool_folder"
    bl_label = "Open spot-image-tool Folder"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        fp = join(bpy.path.abspath('//'), 'tool')
        if not exists(fp):
            mess = f'{fp} not found'
            self.report({'WARNING'}, mess)
            return {'CANCELLED'}
        
        open_folder(fp)
        self.report({'INFO'}, 'Gmic tool folder opened')
        return {'FINISHED'}


##  base panel
class GMICOLOR_PT_auto_color_panel(bpy.types.Panel):
    bl_label = "GP Colorize"# title
    # bl_parent_id # If set, the panel becomes a sub-panel

    ## bl_options = {'DEFAULT_CLOSED', 'HIDE_HEADER' }# closed by default, collapse the panel and the label
    ## is_popover = False # if ommited
    ## bl_space_type = ['EMPTY', 'VIEW_3D', 'IMAGE_EDITOR', 'NODE_EDITOR', 'SEQUENCE_EDITOR', 'CLIP_EDITOR', 'DOPESHEET_EDITOR', 'GRAPH_EDITOR', 'NLA_EDITOR', 'TEXT_EDITOR', 'CONSOLE', 'INFO', 'TOPBAR', 'STATUSBAR', 'OUTLINER', 'PROPERTIES', 'FILE_BROWSER', 'PREFERENCES'], default 'EMPTY'

    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Gpencil"#name of the tab


    # activating on some context only
    ## bl_context : object, objectmode, mesh_edit, curve_edit, surface_edit, text_edit, armature_edit, mball_edit, lattice_edit, pose_mode, imagepaint, weightpaint, vertexpaint, particlemode
    #bl_context = "objectmode"#render

    #need to be in object mode
    @classmethod
    def poll(cls, context):
        return get_addon_prefs().use_color_tools#(context.object is not None and context.object.type == 'GPENCIL')

    ## draw stuff inside the header (place before main label)
    # def draw_header(self, context):
    #     layout = self.layout
    #     layout.label(text="More text in header")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        prefs = get_addon_prefs()
        if not prefs.gmic_path:
            layout.label(text='Gmic path missing in addon prefs', icon='ERROR')
            return

        if not [l for l in context.object.data.layers if 'spot' in l.info]:
            layout.label(text='Need at least one spots layer !', icon='ERROR')
            layout.label(text='("spot" in name to identify)')
            return
        # row = layout.row()
        layout.prop(context.scene.gpcolor_props, 'res_percentage')
        
        ## render and load frame
        ops = layout.operator("bgmic.propagate_color", icon = 'FILE_IMAGE')
        ops.mode = 'render'
        ops.anim = False
        
        ## render and load anim
        ops = layout.operator("bgmic.propagate_color", text='Gmic propagate animation', icon = 'FILE_MOVIE')
        ops.mode = 'render'
        ops.anim = True

        layout.separator()

        ## Load frame
        ops = layout.operator("bgmic.propagate_color", text='load frame', icon = 'FILE_IMAGE')
        ops.mode = 'load'
        ops.anim = False

        ## Load anim
        ops = layout.operator("bgmic.propagate_color", text='Load animation', icon = 'FILE_MOVIE')
        ops.mode = 'load'
        ops.anim = True
        
        layout.separator()
        layout.operator("bgmic.clear_cam_bg_images")

        ## Open gmic tool location
        layout.operator("bgmic.open_gmic_tool_folder", icon = 'FILE_FOLDER')#, text='Open img tool folder'


## TODO : Add an operator to generate empty alpha image to avoid the pink flashes...
## todo : choose to overwrite or not
## Add button to delete current "frame" (delete image on disk...)


classes = (
GMICOLOR_OT_propagate_spots,
GMICOLOR_OT_clear_cam_bg_images,
GMICOLOR_OT_open_gmic_tool_folder,
GMICOLOR_PT_auto_color_panel,
)

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

def unregister():
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)
