import os, re
from os.path import join, basename, exists, dirname, abspath, splitext
import subprocess
import time, datetime

from ..utils import get_addon_prefs, transfer_value

def get_gmic():
    prefs = get_addon_prefs()
    return prefs.gmic_path
    

## globals
image_exts = ('.png', '.jpg', '.tiff', '.tga', '.jpeg',)
Rnum = re.compile(r'(\d+)(?!.*\d)')

def is_img(fp):
    if splitext(basename(fp))[1].lower() in image_exts:
        return True

def is_img_folder(d):
    ct = 0
    for f in os.listdir(d):
        ct += 1
        #if os.path.isfile(d)
        if not is_img(f):
            #print("not an image:", f)#Dbg
            return 0

    if not ct:#return false if nothing found
        #print('nothing in folder')
        return -1

    ###all files has been evaluated as images
    return 1


def auto_colo(line, colo, out):
    gmic = get_gmic()
    opt = r'fx_colorize_lineart_smart 2,96,0,0,0,1,24,197,0,90.4,1,34.05,22.27,48.96,0.57,6.4,1,0,20,50.18,7.5,0.5,0'
    if not exists(line) or not exists(colo):
        print('one source directory not exists')
        return

    line_l = [join(line, i) for i in os.listdir(line) if i.endswith(('.png', '.jpg'))]
    colo_l = [join(colo, i) for i in os.listdir(colo) if i.endswith(('.png', '.jpg'))]
    line_l.sort()
    colo_l.sort()

    #print(line_l)
    #print(colo_l)

    if len(line_l) != len(colo_l):
        print('lists of line and colo files have not the same lenght')
        return

    outfolder_name = basename(dirname(line)) + '_autocolo'
    outfolder = join(out, outfolder_name)
    if not exists(outfolder):
        os.mkdir(outfolder)

    ct = 0
    for i, f in enumerate(line_l):
        ct += 1
        filename = outfolder_name + '_'+ str(i+101).zfill(4) +'.png'
        outfile = join(outfolder, filename)
        #print('-', f,outfile)
        cmd = '{} {} {} {} -o[1] {}'.format(gmic, line_l[i], colo_l[i], opt, outfile)
        #note on -o[1]
        #this filter output 2 img, original line and the new color, here specify only color (keep your name clean)
        print(cmd)
        os.system(cmd)
        # if ct > 3:return#limiter

    print('Done')


def random_fill(line, out):
    '''gmic command to convert line file to pseudo color out file'''
    gmic = get_gmic()
    #Todo, handle png compression, will be deleted so can be uncompressed
    start_time = time.time()#timer
    opt = r'fx_colorize_lineart_smart 0,100,0,0,0,80,184,0,90.5,2,0,0,0,0,4,1,2'
    cmd = [gmic, line] + opt.split(' ') + ['-o[1]', out]# + ['to_rgba']
    print(cmd)
    subprocess.call(cmd)
    print("elapsed",  time.time() - start_time)#timer

def random_fill_folder(src, dest):
    '''gmic command to convert a line folder to pseudo color output in another folder'''
    # opt = r'fx_colorize_lineart_smart 0,97.2,0,0,0,0,210,213,0,86.5,1,49.41,28.67,37.26,0.29,12.7,0,0,0,0,0,0,0'
    if not exists(src):
        print('source directory not exists')
        return

    line_l = [join(src, i) for i in os.listdir(src) if is_img(i)]
    line_l.sort()

    '''
    outfolder_name = basename(dirname(src)) + '_randomcolo'
    outfolder = join(dest, outfolder_name)
    if not exists(outfolder):
        os.mkdir(outfolder)
    '''

    #ct = 0
    for i, f in enumerate(line_l):
        outfile = join(dest, 'random_fill_' + str(i+101).zfill(4) +'.png')
        random_fill(f, outfile)

    print('Done')

def generate_empty_image(fp):
    '''
    Generate an empty 1x1 pixel full transparent
    width,height,depth,spectrum(channels) 
    doc: https://gmic.eu/tutorial/_input.shtml
    generate a file of ~200byte
    '''
    gmic = get_gmic()
    cmd = [gmic, '1,1,1,4', '-o', fp]
    # cmd = [gmic, '16,9,1,4', '-o', fp] #classic ratio... influence nothing
    subprocess.call(cmd)

def propagate_color(line, spot, out):
    gmic = get_gmic()
    start_time = time.time()#timer
    ## colorize
    opt = ['fx_colorize_lineart', '1,3,0,0.02']

    ## antialias
    aa = ['fx_smooth_antialias', '10,10.1,1.22,0,50,50']# smooth aa
    # aa = ['gcd_anti_alias', '60,0.1,0,0']# basic aa

    ## BG remove
    # vert pur 0,255,0,255 for BG color to delete, rose pale : [234,139,147,255] : [0.822786, 0.258183, 0.291771, 1.000000] #EA8B93
    # 2 first value tolerance(1~3),smoothness(need0)
    del_bg = ['to_rgba', 'replace_color', '3,0,1,255,1,255,255,255,255,0']
    print('gmic: ', gmic)
    # gmic = f'"{gmic}"'
    cmd = [gmic, line, spot] + opt + del_bg + ['-o[1]', out]# + aa 
    print('cmd: ', cmd)
    subprocess.call(cmd)
    print("elapsed",  time.time() - start_time)#timer



def propagate_color_folder(line_fp, spot_fp, outfolder):
    print('Starting', datetime.datetime.now())# print full current date
    start_time = time.time()# get start time

    
    # [gmic_krita_qt]./apply/ v -99 fx_colorize_lineart 1,3,0,0.02
    if not exists(line_fp) or not exists(spot_fp) or not exists(outfolder):
        print(f'''some directories not exists:
        {exists(line_fp)}: {line_fp}
        {exists(spot_fp)}: {spot_fp}
        {exists(outfolder)}: {outfolder}''')
        return 1
    
    lines = sorted([f.path for f in os.scandir(line_fp) if is_img(f.name) and Rnum.search(f.name)])
    spots = sorted([f.path for f in os.scandir(spot_fp) if is_img(f.name) and Rnum.search(f.name)])
    
    # if len(lines) != len(spots):#file number test...valid but disabled for tests
    #     print('lists of line and colo files have not the same lenght')
    #     return

    for l, s in zip(lines, spots):
        lframe = int(Rnum.search(l).group(1))
        sframe = int(Rnum.search(s).group(1))   
        if lframe != sframe:
            print(f'line img has not the same number as spot img {lframe} != {sframe}')
            continue
        
        out = join(outfolder, f'colo_{str(lframe).zfill(3)}.png')
        
        print(f'frame {lframe}')
        propagate_color(l,s,out)

    elapsed_time = time.time() - start_time# seconds
    full_time = str(datetime.timedelta(seconds=elapsed_time))# hh:mm:ss format

    print("elapsed time", elapsed_time)
    print(full_time)